package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.core.ControlPanelAction;
import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.face.events.HumanFaceEvent;
import cz.fidentis.analyst.face.events.HumanFaceListener;
import cz.fidentis.analyst.face.events.MeshChangedEvent;
import cz.fidentis.analyst.face.events.SymmetryPlaneChangedEvent;
import cz.fidentis.analyst.feature.FeaturePoint;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshModel;
import cz.fidentis.analyst.scene.DrawableFace;
import cz.fidentis.analyst.scene.DrawablePlane;
import cz.fidentis.analyst.visitors.mesh.HausdorffDistance;
import cz.fidentis.analyst.visitors.mesh.HausdorffDistance.Strategy;

import java.awt.event.ActionEvent;
import java.util.HashMap;
import java.util.Map;
import javax.swing.JTabbedPane;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Action listener for the manipulation with the symmetry plane.
 * 
 * @author Radek Oslejsek
 */
public class SymmetryAction extends ControlPanelAction implements HumanFaceListener  {

    private final SymmetryPanel controlPanel;
    private final JTabbedPane topControlPanel;
    private int sPlane = 0; // 0 = none, 1 = from mesh, 2 = from FPs

    /**
     * Constructor.
     * 
     * @param canvas OpenGL canvas
     * @param topControlPanel Top component for placing control panels
     */
    public SymmetryAction(Canvas canvas, JTabbedPane topControlPanel) {
        super(canvas, topControlPanel);
        this.topControlPanel = topControlPanel;
        //this.controlPanel = new SymmetryPanel(this);
        this.controlPanel = new SymmetryPanel(this);
        
        controlPanel.setComputeFromFPs(getPrimaryFeaturePoints() != null || getSecondaryFeaturePoints() != null);

        // Place control panel to the topControlPanel
        this.topControlPanel.addTab(controlPanel.getName(), controlPanel.getIcon(), controlPanel);
        this.topControlPanel.addChangeListener(e -> {
            // If the symmetry panel is focused...
            if (((JTabbedPane) e.getSource()).getSelectedComponent() instanceof SymmetryPanel) {
                getCanvas().getScene().setDefaultColors();
                getCanvas().getScene().showSymmetryPlanes();
            } else {
                getCanvas().getScene().hideSymmetryPlanes();
            }
        });

        // Be informed about changes in faces perfomed by other GUI elements
        getPrimaryDrawableFace().getHumanFace().registerListener(this);
        if (getSecondaryDrawableFace() != null) {
            getSecondaryDrawableFace().getHumanFace().registerListener(this);
        }
    }
    
    @Override
    public void actionPerformed(ActionEvent ae) {
        String action = ae.getActionCommand();
        
        switch (action) {
            case SymmetryPanel.ACTION_COMMAND_RECOMPUTE: 
                sPlane = 1;
                recomputeFromMesh(0);
                recomputeFromMesh(1);
                break; 
            case SymmetryPanel.ACTION_COMMAND_COMPUTE_FROM_FPS: 
                sPlane = 2;
                recomputeFromFeaturePoints(0);
                recomputeFromFeaturePoints(1);
                break;     
            default:
                // do nothing
        }
        renderScene();
    }
    
    @Override
    public void acceptEvent(HumanFaceEvent event) {
        if (event instanceof MeshChangedEvent) { // recompute symmetry plane, if necessary
            switch (sPlane) {
                case 0: // none
                    return;
                case 1: // from mesh
                    recomputeFromMesh(event.getFace().equals(getPrimaryDrawableFace().getHumanFace()) ? 0 : 1);
                    break;
                case 2: // from FPs
                    recomputeFromFeaturePoints(event.getFace().equals(getPrimaryDrawableFace().getHumanFace())? 0 : 1);
                    break;
                default:
            }
        }
        if (event instanceof SymmetryPlaneChangedEvent) {
            updatePrecision(event.getFace());
        }
    }
    
    protected void recomputeFromMesh(int index) {
        HumanFace face = getCanvas().getScene().getHumanFace(index);
        if (face != null) {
            Logger log = Logger.measureTime();
            SymmetryEstimator estimator = new SymmetryEstimator(controlPanel.getSymmetryConfig());        
            face.getMeshModel().compute(estimator);
            face.setSymmetryPlane(estimator.getSymmetryPlane());
            log.printDuration("Symmetry plane estimation from mesh for " + face.getShortName());
            setDrawablePlane(face, index);
        }
    }
    
    protected void recomputeFromFeaturePoints(int index) {
        HumanFace face = getCanvas().getScene().getHumanFace(index);
        if (face == null) {
            return;
        }
        
        //FeaturePointTypeProvider.getInstance();
        Map<String, Point3d> processFP = new HashMap<>();
        Vector3d dir = new Vector3d();
        int numDirs = 0;
        Point3d centroid = new Point3d();
        
        for (FeaturePoint fp: face.getFeaturePoints()) {
            Point3d auxCentroid = new Point3d(fp.getPosition());
                auxCentroid.scale(1.0 / face.getFeaturePoints().size());
            if (numDirs == 0) {
                centroid.set(auxCentroid);
            } else {
                
                centroid.add(auxCentroid);
            }

            String code = fp.getFeaturePointType().getCode();
            if (code.endsWith("_R")) {
                String pairCode = code.replaceAll("_R", "_L");
                if (processFP.containsKey(pairCode)) {
                    computeDir(processFP.remove(pairCode), fp.getPosition(), dir, numDirs++);
                } else {
                    processFP.put(code, fp.getPosition());
                }
            } else if (code.endsWith("_L")) {
                String pairCode = code.replaceAll("_L", "_R");
                if (processFP.containsKey(pairCode)) {
                    computeDir(fp.getPosition(), processFP.remove(pairCode), dir, numDirs++);
                } else {
                    processFP.put(code, fp.getPosition());
                }                
            }
        }
        
        if (numDirs == 0) { // no pair found
            return;
        }
        
        dir.scale(1.0 / numDirs);
        dir.normalize();
        
        face.setSymmetryPlane(new Plane(dir, dir.dot(new Vector3d(centroid))));
        setDrawablePlane(face, index);
    }
    
    protected void computeDir(Point3d left, Point3d right, Vector3d dir, int numDirs) {
        Vector3d auxDir = new Vector3d(right);
        auxDir.sub(left);
        auxDir.normalize();
        if (numDirs == 0) {
            dir.set(auxDir);
        } else {
            dir.add(auxDir);
        }
    }

    protected void setDrawablePlane(HumanFace face, int index) {
        DrawablePlane symmetryPlane = new DrawablePlane(
                face.getSymmetryPlaneFacet(),
                face.getSymmetryPlane()
        );
        symmetryPlane.setTransparency(0.5f);
        symmetryPlane.setColor((index == 0) ? DrawableFace.SKIN_COLOR_PRIMARY.darker() : DrawableFace.SKIN_COLOR_SECONDARY.darker());
        getCanvas().getScene().setDrawableSymmetryPlane(index, symmetryPlane);
    }
    
    protected void updatePrecision(HumanFace face) {
        if (face == null) {
            return;
        }
        
        MeshModel clone = new MeshModel(face.getMeshModel());
        
        final Vector3d normal = face.getSymmetryPlane().getNormal();
        for (MeshFacet facet: clone.getFacets()) {
            facet.getVertices().parallelStream().forEach(v -> {
                double d = normal.dot(new Vector3d(v.getPosition()));
                Vector3d shiftDir = new Vector3d(normal);
                shiftDir.scale(-2.0 * d);
                v.getPosition().add(shiftDir);
            });
        }
        
        face.computeKdTree(false);
        HausdorffDistance visitor = new HausdorffDistance(face.getKdTree(), Strategy.POINT_TO_POINT_DISTANCE_ONLY, false, true);
        clone.compute(visitor);
        controlPanel.updateHausdorffDistanceStats(visitor.getStats(), getScene().getHumanFace(0) == face);
    }
}
