package cz.fidentis.analyst.tests;

import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.icp.IcpTransformer;
import cz.fidentis.analyst.icp.NoUndersampling;
import cz.fidentis.analyst.visitors.mesh.HausdorffDistance;
import cz.fidentis.analyst.visitors.mesh.HausdorffDistance.Strategy;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.DoubleSummaryStatistics;
import java.util.List;
import java.util.stream.Collectors;

/**
 * A class for testing the efficiency of batch processing algorithms.
 * The goal of this tool is to create a "ground truth" measurements for other optimization techniques
 * 
 * - All pairs are take one by one from the collection
 * - First (primary) face is stored in k-d tree
 * - Secondary face is superimposed using ICP
 * - Hausdorff distance from the secondary to the primary face is computed using POINT_TO_POINT strategy and absolute distances.
 * 
 * @author Radek Oslejsek
 */
public class BatchSimilarityGroundTruth {
    
    private static final String DATA_DIR = "../../analyst-data-antropologie/_ECA";
    private static final String OUTPUT_FILE = "../../SIMILARITY_GROUND_TRUTH.csv";
    private static final int MAX_SAMPLES = 100;
    
    /**
     * Main method 
     * @param args Input arguments 
     * @throws IOException on IO error
     */
    public static void main(String[] args) throws IOException, ClassNotFoundException, Exception {
        List<Path> faces = Files.list(new File(DATA_DIR).toPath())
                .filter(f -> f.toString().endsWith(".obj"))
                .sorted()
                .limit(MAX_SAMPLES)
                .collect(Collectors.toList());
        
        BufferedWriter bfw = new BufferedWriter(new FileWriter(OUTPUT_FILE));
        bfw.write("SEC FACE;PRI FACE;MIN;MAX;AVG (from SEC to PRI);IPC time (mm:ss,mil);HD time (mm:ss,mil)");
        bfw.newLine();
        
        int counter = 1;
        for (int i = 0; i < faces.size(); i++) {
            HumanFace priFace = new HumanFace(faces.get(i).toFile());
            priFace.computeKdTree(false);
            
            for (int j = 0; j < faces.size(); j++) {
                if (i == j) {
                    continue;
                }
                
                HumanFace secFace = new HumanFace(faces.get(j).toFile());
                
                System.out.println(counter++ + "/" + (faces.size()*faces.size()-faces.size()) + " (" + priFace.getShortName() + "/" + secFace.getShortName() + ")");
                
                long icpTime = System.currentTimeMillis();
                IcpTransformer icp = new IcpTransformer(priFace.getMeshModel(), 50, true, 0.3, new NoUndersampling());
                secFace.getMeshModel().compute(icp, true);
                Duration icpDuration = Duration.ofMillis(System.currentTimeMillis() - icpTime);
                
                long hdTime = System.currentTimeMillis();
                HausdorffDistance hd = new HausdorffDistance(priFace.getKdTree(), Strategy.POINT_TO_POINT, false, true);
                secFace.getMeshModel().compute(hd, true);
                DoubleSummaryStatistics stats = hd.getStats();
                Duration hdDuration = Duration.ofMillis(System.currentTimeMillis() - hdTime);
                
                bfw.write(secFace.getShortName()+";");
                bfw.write(priFace.getShortName()+";");
                bfw.write(String.format("%.20f", stats.getMin())+";");
                bfw.write(String.format("%.20f", stats.getMax())+";");
                bfw.write(String.format("%.20f", stats.getAverage())+";");
                bfw.write(String.format("%02d:%02d,%03d", icpDuration.toMinutesPart(), icpDuration.toSecondsPart(), icpDuration.toMillisPart())+";");
                bfw.write(String.format("%02d:%02d,%03d", hdDuration.toMinutesPart(), hdDuration.toSecondsPart(), hdDuration.toMillisPart())+"");
                bfw.newLine();
                bfw.flush();
            }
        }
        
        bfw.close();        
    }

}
